--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   CallGraph.hs
--
-- Author:
--   Jan Hoffmann (2009)
--
-- Description:
--   To be documented.


module Eval where


import qualified Data.Map as M
import Control.Monad.State.Strict


import RAMLTypes
import Monoid

type Heap = M.Map Location Value

type Stack = M.Map VarId Value

data EvalState = St { heap :: Heap
                    , functionDefs :: FunDefs -- = M.Map FunId ([VarId],Exp)
                    , freeLocs :: [Integer]
                    , resCounters :: [(ResourceMetric,MonoidElem)]
                    }



evaluate e funDefList metrics = 
   do { (v,finalState) <- runStateT (eval M.empty e) startState
      ; v_string <- showVal v (heap finalState)
      ; return (v_string, resCounters finalState)
      }
   where
      startState = St { heap = M.empty
                      , functionDefs = M.fromList funDefList
                      , freeLocs = [1..]
                      , resCounters = map (\m -> (m,(0,0)) ) metrics
                      }

      showVal v h = case v of
           VLoc l ->    do { u <- getVal l h
                           ; case u of
                                VList _ _ -> do { list <- showList u h
                                                ; return $ "[" ++ (tail $ list ) ++ "]"
                                                }
                                VBTree _ _ _ -> do { btree <- showBTree u h 0
                                                   ; return $ "\n" ++ btree
                                                   }
                           }
           VBool b ->   return $ show b
           VInt n ->    return $ show n
           VNull ->     return "Null"
           VTuple vs -> do { rec <- sequence [showVal v h | v <- vs]
                           ; return $ "(" ++ (tail (concatMap (\v -> "," ++ v)  rec)) ++ ")"
                           }
     
      printSpaces n = [ ' ' | x <- [1..n] ]

      showBTree v h n = case v of
           VBTree x t1 t2 -> do { val <- showVal x h
                                ; pt1 <- showBTree t1 h (n + 5)
                                ; pt2 <- showBTree t2 h (n + 5)
                                ; return $ (printSpaces n) ++ "node(" ++ val ++ ",\n" ++ pt1 ++ ",\n" ++ pt2 ++ ")"
                                }
           VLoc l  -> do { u <- getVal l h
                         ; showBTree u h n
                         }
           VNull   -> return $ (printSpaces n) ++ "leaf"


      showList v h = case v of 
           VList v1 v2 -> do { head <- (showVal v1 h)
                             ; tail <- (showList v2 h)
                             ; return $ "," ++ head ++ tail
                             }
           VLoc l ->      do { u <- getVal l h
                             ; showList u h
                             }
           VNull  ->      return ""


      getVal l h = case M.lookup l h of
           Just v ->  return v
           Nothing -> fail $ "Trying to access an empty heap cell. \nProbably you "
                      ++ "reused a variable that has been used in a destructive match (matchD)."




eval stack exp = case exp of
             TETick q _         -> do { count $ KTick q
                                      ; return VNull
                                      }
             TEUnit _           -> do { count KUnit
                                      ; return VNull
                                      }
             TETrue _           -> do { count KBool
                                     ; return $ VBool True
                                     }
             TEFalse _          -> do { count KBool
                                     ; return $ VBool False
                                     }
             TENum Normal n _  -> do { count KNat
                                     ; return $ VInt n
                                     }
             TEApp _ f e _     -> do { v <- eval stack e
                                     ; count KApp1
                                     ; (typed_xs,e_f) <- getFunDef f
                                     ; let xs = map fst typed_xs
                                     ; let bindings = case xs of
                                             [x] -> [(x,v)]
                                             _   -> case v of
                                                      VTuple vs -> zip xs vs
                                                      _ -> fail $ "Expecting tuple"
                                                               ++ " and found " ++ (show v) ++ "."
                                     ; v' <- eval (M.fromList bindings) e_f
                                     ; count KApp2
                                     ; return v'
                                     }
             TEVar x _          -> do { count KVar
                                     ; return $ stack M.! x
                                     }
             TEUnaryOp op e _  -> do { v <- eval stack e
                                     ; count $ KUnary op
                                     ; apply1 op v
                                     }
             TEBinOp op e1 e2 _ -> do { v1 <- eval stack e1
                                     ; v2 <- eval stack e2
                                     ; count $ KBin op
                                     ; apply op v1 v2
                                     }
             TESucc Normal e _ -> do { v <- eval stack e
                                     ; count KSucc
                                     ; case v of
                                         VInt n | n >= 0 -> return $ VInt (n+1)
                                         _      -> fail $ "Expected natural number but found "
                                                        ++ (show v)  ++"."
                                     }
             TEToInt e _       -> do { v <- eval stack e
                                     ; count KToInt
                                     ; case v of
                                         VInt n | n >= 0 -> return $ v
                                         _      -> fail $ "Expected natural number but found "
                                                        ++ (show v)  ++"."
                                     }
             TEMatchN e e1 x e2 t -> 
                                  do { v <- eval stack e
                                     ; case v of
                                          VInt 0 -> 
                                             do { count $ KMatchZ1
                                                ; r <- eval stack e1
                                                ; count $ KMatchZ2
                                                ; return r
                                                }
                                          VInt n -> 
                                             do { count $ KMatchS1
                                                ; let stack' = M.insert x (VInt $ n - 1) stack
                                                ; r <- eval stack' e2
                                                ; count $ KMatchS2
                                                ; return r
                                                }
                                          _ -> fail $ "Expected natural number but found "
                                                        ++ (show v)  ++".\n In expression: " 
                                                        ++ (show e) ++ "."
                                     }
             TELet Normal x e1 e2 _ -> do { count KLet1
                                          ; v1 <- eval stack e1
                                          ; count KLet2
                                          ; let stack' = M.insert x v1 stack
                                          ; v2 <- eval stack' e2
                                          ; count KLet3
                                          ; return v2
                                          }
             TECond e e1 e2 _   -> do { VBool b <- eval stack e
                                     ; if b then
                                         do { count KConT1
                                            ; v <- eval stack e1
                                            ; count KConT2
                                            ; return v
                                            }
                                       else
                                         do { count KConF1
                                            ; v <- eval stack e2
                                            ; count KConF2
                                            ; return v
                                            }
                                     }
             TENil Normal _    -> do { count KNil
                                     ; return VNull
                                     }
             TECons Normal e_h e_t _   -> do { v_h <- eval stack e_h
                                             ; v_t <- eval stack e_t
                                             ; v <- store $ VList v_h v_t
                                             ; count $ KCons (size v_h)
                                             ; return v
                                             }
             TEMatchL matchType e e1 x_h x_t e2 _ -> 
                                  do { v <- eval stack e
                                     ; case v of
                                          VNull -> 
                                             do { count $ KMatchN1 matchType
                                                ; r <- eval stack e1
                                                ; count $ KMatchN2 matchType
                                                ; return r
                                                }
                                          (VLoc l) -> 
                                             do { VList v_h v_t <- getValue l
                                                ; count $ KMatchC1 matchType (size v_h)
                                                ; when (matchType == Dest) (free $ VLoc l)
                                                ; let stack' = foldl (\st (x,v) -> M.insert x v st) 
                                                                     stack [(x_h,v_h),(x_t,v_t)]
                                                ; r <- eval stack' e2
                                                ; count $ KMatchC2 matchType
                                                ; return r
                                                }
                                          _ -> fail $ (show v) ++ "\nexp: " ++ (show e)
                                     }
             TETuple Normal es _ -> do { vs <- sequence [eval stack e | e <- es ]
                                       ; count $ KTuple (length vs)
                                       ; return (VTuple vs)
                                       }
             TEMatchT mt e xs e' _ -> do { (VTuple vs) <- eval stack e
                                     ; count $ KMatchT1 mt  --shouldn't it be KMatchT1 length?
                                     ; let stack' = foldl (\st (x,v) -> M.insert x v st) stack (zip xs vs)
                                     ; v <- eval stack' e'
                                     ; count $ KMatchT2 mt
                                     ; return v
                                     }
             TELeaf Normal _   -> do { count KLeaf
                                     ; return VNull
                                     }
             TENodeB Normal e1 e2 e3 _ -> do { val <- eval stack e1
                                             ; t1 <- eval stack e2
                                             ; t2 <- eval stack e3
                                             ; v <- store $ VBTree val t1 t2
                                             ; count $ KNodeB (size val)
                                             ; return v
                                             }
             TEMatchB matchType e e1 x x1 x2 e2 _ -> 
                                  do { v <- eval stack e
                                     ; case v of
                                          VNull -> 
                                             do { count $ KMatchBL1 matchType
                                                ; r <- eval stack e1
                                                ; count $ KMatchBL2 matchType
                                                ; return r
                                                }
                                          (VLoc l) -> 
                                             do { VBTree v v1 v2 <- getValue l
                                                ; count $ KMatchBN1 matchType (size v)
                                                ; when (matchType == Dest) (free $ VLoc l)
                                                ; let stack' = foldl (\st (x,v) -> M.insert x v st) 
                                                                     stack [(x,v),(x1,v1),(x2,v2)]
                                                ; r <- eval stack' e2
                                                ; count $ KMatchBN2 matchType
                                                ; return r
                                                }
                                     }
             _ -> fail $ "Trying to evaluate an expression with Share or Free."

   where
      apply1 op v = case (op,v) of
                      (UMinus, VInt n) -> return $ VInt (-n)
                      (UPlus, VInt n) -> return $ VInt n
                      (UNot, VBool b) -> return $ VBool (not b)
                      _ -> fail $ "Trying to apply unary operator " ++ (show op) 
                           ++ " to value " ++ (show v) ++ "."


      apply op v1 v2 = case (v1,v2) of
            (VInt x, VInt y) -> if op `elem` [Add,Sub,Mult,Mod,Div] 
                                   then let f = case op of
                                                  Add -> (+)
                                                  Sub -> (-)
                                                  Mult -> (*)
                                                  Mod -> mod
                                                  Div -> div
                                        in return $ VInt (f x y)
                                    else let f = case op of
                                                   Eq -> (==)
                                                   LessEq -> (<=)
                                                   GreaterEq -> (>=)
                                                   Less -> (<)
                                                   Greater -> (>)
                                         in
                                         return $ VBool (f x y)
            (VBool x, VBool y) -> let f = case op of
                                            And -> (&&)
                                            Or -> (||)
                                            Eq -> (==)
                                            _   -> fail $ "Trying to apply operator "
                                                        ++ (show op) ++ " to Booleans."
                                  in return $ VBool (f x y)
            (VNull, VNull) -> case op of
                                Eq -> return $ VBool True
                                _   -> fail $ "Trying to apply operator "
                                             ++ (show op) ++ " to Units."
      
            _ -> fail $ "Trying to apply operator "
                          ++ (show op) ++ " to the values " ++ (show v1) ++ " and " ++ (show v2) ++"."
      
      
      store v = do { state@(St {heap = h, freeLocs = l:locs}) <- get
                   ; let h' = M.insert l v h
                   ; put state{heap = h', freeLocs = locs}
                   ; return $ VLoc l
                   }
      
      getValue l = do { St{ heap = h } <- get
                      ; case M.lookup l h of
                          Just v -> return v
                          Nothing -> fail $ "Trying to access an empty heap cell. Probably you "
                                        ++ "reused a variable that has been used in a destructive match (matchD)."
                      }
      
      free (VLoc l) = do { state@(St {heap = h}) <- get 
                         ; let h' = M.delete l h
                         ; put state{heap = h'}
                         }
      
      count k = do { state <- get
                   ; let counters' = map (\(m@(ResourceMetric _ f),c) -> (m,c `dot` f(k)) ) (resCounters state)
                   ; put state{resCounters = counters'}
                   }
      
      getFunDef f = do { state@(St {functionDefs = funDefs}) <- get 
                       ; return $ funDefs M.! f
                       }
