--------------------------------
-------Resource Aware ML--------
--------------------------------
--
-- File:
--   Main.hs
--
-- Author:
--   Jan Hoffmann (2009)
--
-- Description: 
--   The command line/ghci interface for RAML.
--
--   There are two main functions.  See below for a detailed
--   documentation.
--
--   main: is called from command line and should be used from 
--         command line only.
--
--   analyseFile: can be used to analyze RAML programs within ghci.
--
--   The function obcetivesFromFun is also important as it defines how
--   the lp_solver defines an optimal solution.  It should for example
--   be the case, that a it is worth to have a type with a large
--   linear potential then with large constant potential.


module Main where

import System.Environment (getArgs)
import Control.Monad.Error -- To use Either as an error monad
import qualified Data.Map as M
import Data.List
import Data.Maybe (fromJust)
import qualified Data.Time.Clock as Clock

import Parser
import qualified Clp
import qualified LPSolve
import RAMLTypes
import Eval
import SimpleTypes
import ATypes
import Analyse
import Transform
import MultivariatePol
import MBinomial
import MemoryTransform


-- Current version:
version = 0.7
year = 2012
month = "March"

-- parseAndType
--   :: FilePath
--      -> (TExp -> [([Char], [(VarId, Type)], TExp)] -> IO ())
--      -> IO ()

parseAndType fName cont =
  do { inputProg <- readFile fName
     ; case parseProgram fName inputProg of
         Left err -> putStrLn $ "Parse error at " ++ err
         Right (decs,e_main) -> 
             case analyseDecs decs of
               Left err -> putStrLn $ "Well-formed error: " ++ err
               Right funList -> 
                   case typeProgram funList e_main of
                     Left err -> putStrLn $ "Type error: " ++ err
                     Right (typedMain,typedFunList) -> cont typedMain typedFunList
     }


evaluateFile :: FilePath -> IO ()
-- Takes a file name,
-- types the input program and
-- evaluates the main expression.
-- It prints the resulting value and the resource consumption
-- according to the resource metrics that are given in the first line.
evaluateFile fName =
    let metrics = [heapSpace, evalSteps, ticks] in
    parseAndType fName ( \typedMain typedFunList -> 
    case evaluate typedMain (concat typedFunList) metrics of
      Left err -> putStrLn $ "Evaluation error: " ++ err
      Right (v,counters) -> do {
        ; putStrLn $ "main : " ++ show (typeOf typedMain)
        ; putStrLn $ ""
        ; putStrLn $ "main = " ++ v
        ; putStrLn $ ""
        ; putStrLn $ "Resource Usage:"
        ; sequence_ [ putStrLn ("     " ++ (show (fromRational q :: Double)) ++ " " ++ name) 
                    | (ResourceMetric name _, (q,q')) <- counters ]
      } )



ramlOption o =   o == "--verbose"
              || "--tempdir" `isPrefixOf` o
              || o == "--time"
              || o == "--withclone"
              || o == "--clp"
              || o == "--lp_solve"
              || o == "--keepfiles"
-- The options that can be used on the command line.  For each
-- options, a boolean function that states which strings are valid
-- options of that type.


main :: IO ()
-- Parses the command line options and calls the functions analyse or
-- evaluateFile with the given parameters.
-- main should be used from command line only.
main = do { args <- getArgs
          ; let (options,nonOptions) = partition (\a -> "--" `isPrefixOf` a) args
          ; sequence_ [when (not (ramlOption o)) (error $ o ++ " is not a valid option.") | o <- options ]
          ; startTime <- Clock.getCurrentTime
          ; let outputMode = if "--verbose" `elem` options then Verbose else Short
          ; let clonetransform = "--withclone" `elem` options
          ; let tempFilesFlag = if "--keepfiles" `elem` options then KeepFiles else DeleteFiles
          ; let tempDir =  case find (isPrefixOf "--tempdir=") options of
                             Nothing  -> "."
                             Just tmp -> fromJust (stripPrefix "--tempdir=" tmp)
          ; when (length nonOptions < 1) (error "Usage: OPTIONS ACTION ARGUMENTS FILE")
          ; let ([action],args') = splitAt 1 nonOptions
          ; putStrLn $ "Resource Aware ML, Version " ++ (show version) ++", "++ month ++ " "++(show year) ++"."
          ; putStrLn $ ""
          ; case action of
                "analyse" ->
                    do { when (length args' /= 3) (error "Usage: analyse METRIC DEGREE FILE")
                       ; let ([metricS,degree,fName],_) = splitAt 3 args'
                       ; let maxDeg = read degree :: Int
                       ; when (maxDeg < 1) (error $ "Usage: analyse METRIC DEGREE FILE. The maximal "
                                                 ++ "degree must be greater then 0.")
                       ; let solveFun = if "--lp_solve" `elem` options 
                                          then LPSolve.solve tempFilesFlag
                                          else Clp.solve tempFilesFlag
                       ; let metric = case metricS of
                              "heap-space" -> heapSpace
                              "eval-steps" -> evalSteps
                              "overflow"   -> overflow
                              "cost-free"  -> costFree
                              "ticks"      -> ticks
                              _            ->  error $ "Usage: analyse METRIC DEGREE FILE. The metric "
                                                    ++ "can be 'heap-space', 'eval-steps', 'overflow'"
                                                    ++ " or 'cost-free'."
                       ; putStrLn "Analyzing ..."
                       ; putStrLn ""
                       ; analyse fName metric maxDeg tempDir outputMode clonetransform solveFun
                       }
                "evaluate" -> 
                    do { when (length args' /= 1) (error "Usage: evaluate FILE")
                       ; let fName = head args'
                       ; putStrLn "Evaluating ..."
                       ; putStrLn ""
                       ; evaluateFile fName  
                       }
                _ | action `elem` ["version","about","info"] -> 
                    putStrLn $ "Resource Aware ML, " ++ "Version " ++ (show version) ++ ", " ++ (show year)
                "RAMLer" -> printRAMLer
                _ -> error $ "Usage: OPTIONS ACTION ARGUMENTS FILE.  Action " ++ action ++ " unknown."
          ; if "--time" `elem` options then printTime startTime else return ()
          }
  where
     printTime startTime =
      do { endTime <- Clock.getCurrentTime
         ; let usedTime = Clock.diffUTCTime endTime startTime 
         ; putStrLn ""
         ; putStrLn "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -"
         ; putStrLn ""
         ; putStrLn $ "Run time of the prototype: " ++ (show usedTime)
         }




analyseFile :: FilePath -> ResourceMetric -> Int -> IO ()
-- Takes a file name, a resource metric (see RAMLTypes.hs), and a
-- maximal degree.  Then analyses the file with respect to the
-- metric and the degree.
-- The current directory is used as temp dir.
-- The used solver is Clp.
analyseFile fName metric maxDegree =
   analyse fName metric maxDegree "." Verbose False (Clp.solve DeleteFiles)



analyse  :: (Real a, Fractional a) =>
        String
     -> ResourceMetric
     -> Int
     -> FilePath
     -> OutputMode
     -> Bool -- clone transformation
     -> ([Constraint]  -- a solve function
         -> [Objective]
         -> FilePath
         -> IO (Either [Char] (M.Map CVarId a)))
     -> IO ()
-- Takes a file name (of a RAML program), a resource metric, a maximal
-- degree, and a directory for temp files.  Then
-- - checks if the declarations are well-formed
-- - types the program
-- - solves the linear constraints via lp_solve
-- - and prints for each function: a function type and a list of
--   function names and types that have been used to type the
--   function.

analyse fName metric maxDegree tempDir outputMode clonetransform solveFun =
   do { when (maxDegree < 1) (putStrLn $ "Error: The degree must be greater then 0.")
      ; parseAndType fName ( \typedMain typedFunList -> do {
      ; let typedFunLists = normalizeFunDecs typedFunList  -- list is in reverse topological order
      ; let typedMain'    = shareLetNormal [ "^main"++(show i) | i <- [1..] ] typedMain
      ; putStrLn ""
      ; if outputMode == Verbose  then 
            do {
              putStrLn "";
              putStrLn "The following Sharing and Let-normal forms are used";
              putStrLn "---------------------------------------------------\n";
              putStr (foldl (++) "" (map (\(f,(args,exp)) -> "Function " ++ f ++ " with arguments " ++ show args ++ "\n" ++ show exp ++ "\n\n") (concat typedFunLists))) ;
              putStrLn $ "Main expression:\n" ++ (show typedMain');
           } else return ()
      ; let annoFunTypeLists = annotateProgram typedFunLists typedMain' metric maxDegree
      ; sequence_ [ annalyseSCC annoFunTypeList sccNum | (annoFunTypeList,sccNum) <- (zip annoFunTypeLists [1..]) ]
     } ) }
  where
    annalyseSCC (annoFunTypeList,constraints,namedFunAppList) sccNum =
      do { putStrLn $ "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -"
         ; putStrLn $ ""
         ; putStrLn $ "Analyzing SCC " ++ (show sccNum) ++ " of the call graph."
         ; putStrLn $ ""
         ; putStrLn $ (show $ length constraints) ++ " constraints generated."
         ; putStrLn "Solving ... "
         ; let objectiveFun = concatMap objectivesFromFun (map (\(f,tf) -> tf) annoFunTypeList)
         ; lpResult <- solveFun constraints objectiveFun tempDir
         ; case lpResult of
            Left err -> putStrLn $ "Failed to compute the type annotations: " ++ err
            Right solution -> do { printTypes annoFunTypeList TopLevel solution metric outputMode
                                 ; case namedFunAppList of
                                       [] -> return ()
                                       _  -> do { putStrLn "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -"
                                                ; putStrLn ""
                                                ; putStrLn $ "Function applications with user labels in SCC " ++ (show sccNum) ++ ":"
                                                ; printTypes namedFunAppList FunctionCalls solution metric outputMode
                                                }
                                       
                                 }
        }

data PrintMode = TopLevel | FunctionCalls



-- printTypes prints a list of function types.  It needs a mapping
-- from variables to doubles, a metric (just to get the name), and and
-- OutputMode.
printTypes annoFunTypeList printMode solution metric@(ResourceMetric metricName _ ) outputMode =
   do { putStrLn ""
      ; printFunTypes annoFunTypeList
      }
   where
      lenVars = ["n","m","x","y","z","u","v","w","p","q","r","s","a","b","c","d","e","f","g","h","i","j","k"]
                ++ [ "n"++(show i) | i <- [1..] ]

      indent = "       "

      val p = case M.lookup p solution of
                Nothing -> 0.0
                Just s -> s  
              -- Now result type annotations are not part of the objective function.
              -- If a part of the function's result is constantly nil then the respective
              -- annotations are unconstrained.
              -- Was:   error $ "Variable " ++ (show p) ++ " is not part of the solution."
      
      printFunTypes annoFunTypeList = case annoFunTypeList of
         []      -> return ()
         ('_':_, _):fTList -> printFunTypes fTList  -- for tools that need intermediate functions that should not be shown
         (f,t_f):fTList
                 -> let AFunType (AType (tArg,polArg@(MPolynomial fArg)) ) (AType (tRes,polRes)) = t_f in
                    do { putStrLn "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -"
                       ; putStrLn ""
                       ; putStr   $ f ++ ": " ++ (show tArg) ++ " --> " ++  (show tRes)
                       ; putStrLn ""
                       ; putStrLn "Positive annotations of the argument         Positive annotations of the result"
                       ; putStrsInterleave 45 (posAnnos polArg) (posAnnos polRes)
                       ; putStrLn ""
                       ; case printMode of
                            TopLevel -> do { if metric == overflow
                                               then putStrLn $ "All computed natural numbers in the evaluation of " ++ f ++ " are less or equal to:"
                                               else putStrLn $ "The number of "++ metricName ++" consumed by " ++ f ++ " is at most:"
                                           ; putStr   $ indent ++ "   " 
                                           ; putStrLn $ describePol lenVars tArg (M.map (\v -> toRational (val v)) fArg)
                                           ; putStrLn ""
                                           }
                            FunctionCalls -> return ()
--                        ; let unconstRes = unconstrResults polRes
--                        ; when (unconstRes /= [])
--                           (putStrLn $ "WARNING: The annotations for the indices " ++ (show unconstRes) ++ " are unconstrained in the result.\n")
-- Clp outputs only non-zero values
                       ; printFunTypes fTList
       }

      posAnnos (MPolynomial poly) = 
         let posIndices = filter (\i -> (val (poly M.! i)) /= 0 ) (M.keys poly)
             maxLength = foldl (\n i -> max n (length (show i)) ) 0 posIndices
             showIndVal ind = let strI = (show ind) in
                              strI ++ (spaces (max 0 (maxLength - (length strI) )) ) ++ " --> " ++ (show (val (poly M.! ind)))
          in
          map showIndVal posIndices

      unconstrResults (MPolynomial poly) =
          filter isunconstr (M.keys poly)
          where isunconstr i = case M.lookup (poly M.! i) solution of
                                 Nothing -> True
                                 Just _ -> False

      putStrsInterleave columnLength strs1 strs2 =
         case (strs1,strs2) of
           (s1:strs1,s2:strs2) -> do { putStrs s1 s2
                                     ; putStrsInterleave columnLength strs1 strs2
                                     }
           (s1:strs1,[]      ) -> do { putStrs s1 ""
                                     ; putStrsInterleave columnLength strs1 []
                                     }
           ([]      ,s2:strs2) -> do { putStrs "" s2
                                     ; putStrsInterleave columnLength [] strs2
                                     }
           ([]      ,[]      ) -> return ()

        where
          putStrs s1 s2 = putStrLn $ s1 ++ (spaces (max 3 (columnLength - (length s1)) ) ) ++ s2

      spaces n = [ ' ' | i <- [1..n] ]





printRAMLer =
    do{ putStrLn "   \\\\           "
      ; putStrLn "    \\\\_   \\\\    "
      ; putStrLn "     (')   \\\\_  "
      ; putStrLn "    / )=.---(') "
      ; putStrLn "  o( )o( )_-\\_  "
      }




