--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   Sharing.hs
--
-- Author:
--   Klaus Aehlig (2010)
--
-- Description:
--   * compute the coefficients for polynomials obtained
--     by sharing of variables

module Sharing where

import MultivariatePol
import Util
import qualified Data.Map as M



interleave :: [TIndices] -> [TIndices] -> [[(TIndices, TIndices)]]
-- Input:  two lists of indices
-- Output: a list of all possible interleavings where two indices can
--         be at identical positions.  It is handy to distinguish the
--         the origin of the indices by using pairs where TIUNIT means
--         empty.
--
-- Example: interleave [1,2] [3,4] = [[(1,*),(2,*),(*,3),(*,4)]
--                                   ,[(1,*),(*,3),(2,*),(*,4)],[(1,*),(*,3),(*,4),(2,*)]
--                                   ,[(1,*),(*,3),(2,4)],      [(1,*),(2,3),(*,4)]
--                                   ,[(*,3),(1,*),(2,*),(*,4)],[(*,3),(1,*),(*,4),(2,*)]
--                                   ,[(*,3),(1,*),(2,4)],      [(*,3),(*,4),(1,*),(2,*)]
--                                   ,[(*,3),(1,4),(2,*)],      [(1,3),(2,*),(*,4)]
--                                   ,[(1,3),(*,4),(2,*)],      [(1,3),(2,4)]]

interleave  xs     []     = return  (map (\ x -> (x,TIUnit)) xs)
interleave  []     ys     = return  (map (\ y -> (TIUnit,y)) ys)
interleave  (x:xs) (y:ys) = (interleave xs     (y:ys) >>= \l -> [(x,TIUnit):l]) ++
                            (interleave (x:xs) ys     >>= \l -> [(TIUnit,y):l]) ++
                            (interleave xs     ys     >>= \l -> [(x,y):l])


collect :: [(t, Integer)] -> ([t], Integer)
-- Input: A list of pairs of indices and integers
-- Output: A pair where the first component is the list of the first components
--         and the second component is the product of the second components
-- Used in sharingCoeff for tuples and lists.

collect = foldr (\(t,c) -> \(ts,n) -> (t:ts,c*n)) ([],1)


sharingCoeff :: TIndices -> TIndices -> M.Map TIndices Integer
-- Input: Two indices i,j
-- Output: A map k -> c^(i,j)_k such that
--         p_i*p_j = sum_k c^(i,j)_k p_k
--
-- sharingCoeff should be called with indices of the same type only.
-- It is handy to allow one argument to be TIUnit regardless of the other
-- argument.  The sharingCoeffs are then 1, which is equal to p_*.

sharingCoeff a            TIUnit       = M.fromList [(a,1)]

sharingCoeff TIUnit       b            = M.fromList [(b,1)]

sharingCoeff (TITuple xs)(TITuple ys)  = 
    M.fromList 
         (map (\(ts,n) -> (TITuple ts,n))
              (map collect
                   (combine (map M.toList (map2 sharingCoeff xs ys)))))

sharingCoeff (TIList xs)(TIList ys)  = 
    foldl (M.unionWith (+)) M.empty
          (interleave xs ys 
            >>= combine . map (M.toList . uncurry sharingCoeff)
            >>= (\(xs,c) -> return (M.fromList [(TIList xs, c)])) . collect)

